#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains supporting C++ headers to distribute with generated types.
"""

import pathlib
import typing

__version__ = "1.0.0"
"""Version of the c++ support headers."""


def copy_support_headers(target_namespace: typing.List[str],
                         copy_to_folder: pathlib.Path,
                         allow_overwrite: bool) -> typing.List[pathlib.Path]:
    """
    Copy C++ support headers out of the Nunavut package and into the folder specified.

    .. invisible-code-block: python

        from nunavut.lang.cpp.support import copy_support_headers
        import pathlib
        import pytest

    .. code-block:: python

        # copied with contain headers like 'test_out_dir/nunavut/support/serialization.hpp'
        copied = copy_support_headers(['nunavut', 'support'], gen_paths.out_dir, True)

        # To prevent overwrites set the allow_overwrite parameter to False
        with pytest.raises(PermissionError):
            copied = copy_support_headers(['nunavut', 'support'], gen_paths.out_dir, False)

    .. invisible-code-block: python

        assert len(copied) > 0
        for header in copied:
            assert(header.is_file)

    :param pathlib.Path copy_to_folder: The folder to copy the headers into.
    :param bool allow_overwrite: If True then this method will copy the support files over
        existing files of the same name.
    :return: A list of paths to the copied headers.
    :raises: PermissionError if :attr:`allow_overwrite` is False and the file exists.
    """
    import pkg_resources
    import shutil
    copied = []  # type: typing.List[pathlib.Path]
    resources = list_support_headers([])
    try:
        copy_to_folder.mkdir()
    except FileExistsError:
        pass
    namespaced_path = copy_to_folder
    for namespace_part in target_namespace:
        namespaced_path = namespaced_path / pathlib.Path(namespace_part)
        try:
            namespaced_path.mkdir()
        except FileExistsError:
            pass
    for resource in resources:
        target = namespaced_path / pathlib.Path(resource)
        if not allow_overwrite and target.exists():
            raise PermissionError('{} exists. Refusing to overwrite.'.format(str(target)))
        shutil.copy(pkg_resources.resource_filename(__name__, str(resource)), str(namespaced_path))
        copied.append(target)
    return copied


def list_support_headers(target_namespace: typing.List[str]) -> typing.List[pathlib.Path]:
    """
    Get a list of C++ support headers embedded in this package.

    .. invisible-code-block: python

        from nunavut.lang.cpp.support import list_support_headers
        import pathlib

    .. code-block:: python

        paths = list_support_headers([])
        for path in paths:
            assert pathlib.Path('') == path.parent

        paths = list_support_headers(['nunavut'])
        for path in paths:
            assert pathlib.Path('nunavut') == path.parent

        paths = list_support_headers(['nunavut', 'support'])
        for path in paths:
            assert pathlib.Path('nunavut') / pathlib.Path('support') == path.parent

    :return: A list of C++ support header resources.
    """
    import pkg_resources
    namespace_path = pathlib.Path('')
    for namespace_part in target_namespace:
        namespace_path = namespace_path / pathlib.Path(namespace_part)
    headers = []  # type: typing.List[pathlib.Path]
    resources = [r for r in pkg_resources.resource_listdir(__name__, '.') if r.endswith('.hpp')]
    for resource in resources:
        headers.append(namespace_path / pathlib.Path(resource))
    return headers



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Filters for generating C++. All filters in this
    module will be available in the template's global namespace as ``cpp``.
"""

import functools
import io
import pathlib
import re
import typing

import pydsdl

from ...generators import AbstractGenerator
from ...lang import Language
from ...templates import (template_language_filter,
                          template_language_list_filter)
from ..c import C_RESERVED_PATTERNS, VariableNameEncoder, _CFit
from ._utilities import IncludeGenerator

CPP_RESERVED_PATTERNS = frozenset([*C_RESERVED_PATTERNS])

CPP_NO_DOUBLE_DASH_RULE = re.compile(r'(__)')


class SupportGenerator(AbstractGenerator):
    """
    Copy C++ support types to the :func:`support_output_folder <nunavut.Namespace.get_support_output_folder()>`.
    This class name is expected by the :func:`nunavut.generators.create_support_generator()` method.

    .. invisible-code-block: python

        import pathlib
        from unittest.mock import MagicMock, NonCallableMagicMock
        from nunavut import YesNoDefault
        from nunavut.lang import LanguageContext
        from nunavut.lang.cpp import SupportGenerator

        language_context = LanguageContext()
        namespace = NonCallableMagicMock()
        namespace.get_language_context = MagicMock()
        namespace.get_language_context.return_value = language_context
        namespace.get_support_output_folder = MagicMock()
        namespace.get_support_output_folder.return_value = gen_paths.out_dir

    .. code-block:: python

        generator = SupportGenerator(namespace, YesNoDefault.DEFAULT)
        assert 0 == generator.generate_all()

    """
    def generate_all(self,
                     is_dryrun: bool = False,
                     allow_overwrite: bool = True) \
            -> int:
        from .support import copy_support_headers
        language = self.namespace.get_language_context().get_language(__package__)

        output_folder = pathlib.Path(self.namespace.get_support_output_folder())
        copy_support_headers(language.support_namespace, output_folder, allow_overwrite)

        return 0


@template_language_filter(__name__)
def filter_id(language: Language,
              instance: typing.Any) -> str:
    """
    Filter that produces a valid C and/or C++ identifier for a given object. The encoding may not
    be reversable.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_id


    .. code-block:: python

        # Given
        I = 'I like c++'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_like_cZX002BZX002B'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = 'if'

        # and
        template = '{{ I | id }}'

        # then
        rendered = '_if'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = '_Reserved'

        # and
        template = '{{ I | id }}'

        # then
        rendered = '_reserved'

    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)

    :param any instance:        Any object or data that either has a name property or can be converted
                                to a string.
    :returns: A token that is a valid identifier for C and C++, is not a reserved keyword, and is transformed
              in a deterministic manner based on the provided instance.
    """
    if hasattr(instance, 'name'):
        raw_name = str(instance.name)  # type: str
    else:
        raw_name = str(instance)

    vne = VariableNameEncoder(language.stropping_prefix, language.stropping_suffix, language.encoding_prefix)
    reserved_identifiers = frozenset(language.get_reserved_identifiers())
    out = vne.strop(raw_name,
                    reserved_identifiers,
                    CPP_RESERVED_PATTERNS)
    return CPP_NO_DOUBLE_DASH_RULE.sub('_' + vne.encode_character('_'), out)


@template_language_filter(__name__)
def filter_open_namespace(language: Language,
                          full_namespace: str,
                          bracket_on_next_line: bool = True,
                          linesep: str = '\n') -> str:
    """
    Emits c++ opening namspace syntax parsed from a pydsdl "full_namespace",
    dot-seperated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_open_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | open_namespace }}'

        # then
        rendered = '''namespace uavcan
        {
        namespace foo
        {
        '''

    .. invisible-code-block: python

        jinja_filter_tester(filter_open_namespace, template, rendered, 'cpp', T=T)

    :param str full_namespace: A dot-seperated namespace string.
    :param bool bracket_on_next_line: If True (the default) then the opening
        brackets are placed on a newline after the namespace keyword.
    :param str linesep: The line-seperator to use when emitting new lines.
                        By default this is ``\\n``.

    :returns: C++ namespace declarations with opening brackets.
    """

    with io.StringIO() as content:
        for name in full_namespace.split('.'):
            content.write('namespace ')
            if language.enable_stropping:
                content.write(filter_id(language, name))
            else:
                content.write(name)
            if bracket_on_next_line:
                content.write(linesep)
            else:
                content.write(' ')
            content.write('{')
            content.write(linesep)
        return content.getvalue()


@template_language_filter(__name__)
def filter_close_namespace(language: Language,
                           full_namespace: str,
                           omit_comments: bool = False,
                           linesep: str = '\n') -> str:
    """
    Emits c++ closing namspace syntax parsed from a pydsdl "full_namespace",
    dot-seperated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_close_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | close_namespace }}'

        # then
        rendered = '''} // namespace foo
        } // namespace uavcan
        '''

    .. invisible-code-block: python

        jinja_filter_tester(filter_close_namespace, template, rendered, 'cpp', T=T)


    :param str full_namespace: A dot-seperated namespace string.
    :param bool omit_comments: If True then the comments following the closing
                               bracket are omitted.
    :param str linesep: The line-seperator to use when emitting new lines.
                        By default this is ``\\n``

    :returns: C++ namespace declarations with opening brackets.
    """
    with io.StringIO() as content:
        for name in reversed(full_namespace.split('.')):
            content.write('}')
            if not omit_comments:
                content.write(' // namespace ')
                if language.enable_stropping:
                    content.write(filter_id(language, name))
                else:
                    content.write(name)
            content.write(linesep)
        return content.getvalue()


@template_language_filter(__name__)
def filter_full_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is the full namespace, typename, major, and minor version for a given composite type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_full_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_obj = MagicMock()
        my_obj.parent_service = None
        my_obj.version = MagicMock()

    .. code-block:: python

        # Given a type with illegal characters for C++
        my_obj.full_name = 'any.int.2Foo'
        my_obj.version.major = 1
        my_obj.version.minor = 2

        # and
        template = '{{ my_obj | full_reference_name }}'

        # then, with stropping enabled
        rendered = 'any::_int::_2Foo_1_2'

    .. invisible-code-block: python

        my_obj.short_name = my_obj.full_name.split('.')[-1]
        jinja_filter_tester(filter_full_reference_name, template, rendered, 'cpp', my_obj=my_obj)

        my_obj = MagicMock()
        my_obj.version = MagicMock()
        my_obj.parent_service = None

    .. invisible-code-block: python

        my_obj = MagicMock(spec=pydsdl.CompositeType)
        my_obj.version = MagicMock()
        my_service = MagicMock(spec=pydsdl.ServiceType)
        my_service.parent_service = None
        my_service.version = MagicMock()
        my_service.attributes = { 'Request': my_obj }
        my_obj.parent_service = my_service

    Note that for service types

    .. code-block:: python

        # Given a service type
        my_service.full_name = 'my.Service'
        my_service.version.major = 1
        my_service.version.minor = 8

        # and
        template = '{{ my_service.attributes["Request"] | full_reference_name }}'

        # then
        rendered = 'my::Service_1_8::Request'

    .. invisible-code-block: python

        my_service.short_name = my_service.full_name.split('.')[-1]
        my_obj.short_name = 'Request'
        my_obj.full_name = my_service.full_name + '.' + my_obj.short_name

        jinja_filter_tester(filter_full_reference_name, template, rendered, 'cpp', my_service=my_service)

    :param pydsdl.CompositeType t: The DSDL type to get the fully-resolved reference name for.
    """
    ns_parts = t.full_name.split('.')
    if language.enable_stropping:
        ns = list(map(functools.partial(filter_id, language), ns_parts[:-1]))
    else:
        ns = ns_parts[:-1]

    if t.parent_service is not None:
        assert len(ns) > 0  # Well-formed DSDL will never have a request or response type that isn't nested.
        ns = ns[:-1] + [filter_short_reference_name(language, t.parent_service)]

    full_path = ns + [filter_short_reference_name(language, t)]
    return '::'.join(full_path)


@template_language_filter(__name__)
def filter_short_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is a shorted version of the full reference name. This type is unique only within its
    namespace.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_short_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

    .. code-block:: python

        # Given a type with illegal C++ characters
        my_type.short_name = '2Foo'
        my_type.version.major = 1
        my_type.version.minor = 2

        # and
        template = '{{ my_type | short_reference_name }}'

        # then, with stropping enabled
        rendered = '_2Foo_1_2'

    .. invisible-code-block: python

        jinja_filter_tester(filter_short_reference_name, template, rendered, 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

    :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
    """
    if t.parent_service is None:
        short_name = '{short}_{major}_{minor}'.format(short=t.short_name, major=t.version.major, minor=t.version.minor)
    else:
        short_name = t.short_name
    if language.enable_stropping:
        return filter_id(language, short_name)
    else:
        return short_name


@template_language_list_filter(__name__)
def filter_includes(language: Language,
                    t: pydsdl.CompositeType,
                    sort: bool = True) -> typing.List[str]:
    """
    Returns a list of all include paths for a given type.

    :param pydsdl.CompositeType t: The type to scan for dependencies.
    :param bool sort: If true the returned list will be sorted.
    :return: a list of include headers needed for a given type.
    """

    include_gen = IncludeGenerator(language,
                                   t,
                                   filter_id,
                                   filter_short_reference_name)
    return include_gen.generate_include_filepart_list(language.extension, sort)


@template_language_filter(__name__)
def filter_declaration(language: Language,
                       instance: pydsdl.Any,
                       use_standard_types: typing.Optional[bool] = None) -> str:
    """
    Emit a declaration statement for the given instance.
    """
    if use_standard_types is None:
        use_standard_types = bool(language.get_config_value('use_standard_types'))
    if isinstance(instance, pydsdl.PrimitiveType) or isinstance(instance, pydsdl.VoidType):
        return filter_type_from_primitive(language, instance, use_standard_types)
    elif isinstance(instance, pydsdl.VariableLengthArrayType):
        return 'std::vector<{}>'.format(filter_declaration(language, instance.element_type, use_standard_types))
    elif isinstance(instance, pydsdl.ArrayType):
        return 'std::Array<{}>'.format(filter_declaration(language, instance.element_type, use_standard_types))
    else:
        return filter_full_reference_name(language, instance)


@template_language_filter(__name__)
def filter_definition_begin(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the start of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_begin
        from unittest.mock import MagicMock
        import pytest
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_begin,
                                '{{ my_type | definition_begin }}',
                                '',
                                'cpp',
                                my_type=MagicMock())

    .. code-block:: python

        # Given a pydsdl.CompositeType "my_type":
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        # and
        template = '{{ my_type | definition_begin }}'

        # then
        rendered = 'struct Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_type=my_type)

        my_union_type = MagicMock(spec=pydsdl.UnionType)
        my_union_type.version = MagicMock()
        my_union_type.parent_service = None

    .. code-block:: python

        # Also, given a pydsdl.UnionType "my_union_type":
        my_union_type.short_name = 'Foo'
        my_union_type.version.major = 1
        my_union_type.version.minor = 0

        # and
        union_template = '{{ my_union_type | definition_begin }}'

        # then
        rendered = 'union Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, union_template, rendered, 'cpp', my_union_type=my_union_type)

        my_service_type = MagicMock(spec=pydsdl.ServiceType)
        my_service_type.version = MagicMock()
        my_service_type.parent_service = None

    .. code-block:: python

        # Finally, given a pydsdl.Servicetype "my_service_type":
        my_service_type.short_name = 'Foo'
        my_service_type.version.major = 1
        my_service_type.version.minor = 0

        # and
        template = '{{ my_service_type | definition_begin }}'

        # then
        rendered = 'namespace Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_service_type=my_service_type)

    """
    short_name = filter_short_reference_name(language, instance)
    if isinstance(instance, pydsdl.StructureType):
        return 'struct {}'.format(short_name)
    elif isinstance(instance, pydsdl.UnionType):
        return 'union {}'.format(short_name)
    elif isinstance(instance, pydsdl.ServiceType):
        return 'namespace {}'.format(short_name)
    else:
        raise ValueError('{} types cannot be redefined.'.format(type(instance).__name__))


@template_language_filter(__name__)
def filter_definition_end(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the end of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_end
        from unittest.mock import MagicMock
        import pytest
        import pydsdl


        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', '', 'cpp', my_type=MagicMock())

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.UnionType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.ServiceType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end,
                            '{{ my_type | definition_end }}',
                            ' // namespace Foo_1_0',
                            'cpp',
                            my_type=my_type)

    """
    if isinstance(instance, pydsdl.StructureType) or isinstance(instance, pydsdl.UnionType):
        return ';'
    elif isinstance(instance, pydsdl.ServiceType):
        return ' // namespace {}'.format(filter_short_reference_name(language, instance))
    else:
        raise ValueError('{} types cannot be redefined.'.format(type(instance).__name__))


@template_language_filter(__name__)
def filter_type_from_primitive(language: Language,
                               value: pydsdl.PrimitiveType,
                               use_standard_types: typing.Optional[bool] = None) -> str:
    if use_standard_types is None:
        use_standard_types = bool(language.get_config_value('use_standard_types'))
    return _CFit.get_best_fit(value.bit_length).to_c_type(value, use_standard_types)


def filter_to_namespace_qualifier(namespace_list: typing.List[str]) -> str:
    """
    Converts a list of namespace names into a qualifer string. For example:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_namespace_qualifier
        import pydsdl


    .. code-block:: python

        my_namespace = ['foo', 'bar']
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'foo::bar::myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    This filter gracefully handles empty namespace lists:

    .. code-block:: python

        my_namespace = []
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    """
    if namespace_list is None or len(namespace_list) == 0:
        return ''
    else:
        return '::'.join(namespace_list) + '::'



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Language-specific support in nunavut.

This package contains modules that provide specific support for generating
source for various languages using templates.
"""
import configparser
import functools
import importlib
import inspect
import logging
import pathlib
import typing

logger = logging.getLogger(__name__)


class Language:
    """
    Facilities for generating source code for a specific language.

    :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.
    :param configparser.ConfigParser config: The parser to load language properties from.
    :param bool omit_serialization_support:  The value to set for the :func:`omit_serialization_support` property
                                             for this language.
    """

    @classmethod
    def _find_filters_for_language(cls, language_name: str) -> typing.Mapping[str, typing.Callable]:
        filter_map = dict()  # type: typing.Dict[str, typing.Callable]
        lang_module = importlib.import_module('nunavut.lang.{}'.format(language_name))
        filters = inspect.getmembers(lang_module, inspect.isfunction)
        for function_tuple in filters:
            function_name = function_tuple[0]
            if len(function_name) > 7 and function_name[0:7] == "filter_":
                filter_map[function_name[7:]] = function_tuple[1]
                logging.debug("Adding filter {} for language {}".format(function_name[7:],
                                                                        language_name))
        return filter_map

    def __init__(self,
                 language_name: str,
                 config: configparser.ConfigParser,
                 omit_serialization_support: bool):
        self._language_name = language_name
        self._section = 'nunavut.lang.{}'.format(language_name)
        self._filters = self._find_filters_for_language(language_name)
        self._config = config
        self._omit_serialization_support = omit_serialization_support
        self._module = None  # type: typing.Optional[typing.Any]

    @property
    def extension(self) -> str:
        """
        The extension to use for files generated in this language.
        """
        return self._config.get(self._section, 'extension')

    @property
    def namespace_output_stem(self) -> typing.Optional[str]:
        """
        The name of a namespace file for this language.
        """
        return self._config.get(self._section, 'namespace_file_stem')

    @property
    def name(self) -> str:
        """
        The name of the language used by the :mod:`nunavut.lang` module.
        """
        return self._language_name

    @property
    def stropping_prefix(self) -> str:
        """
        The string to prepend to an identifier when stropping.
        """
        return self._config.get(self._section, 'stropping_prefix')

    @property
    def stropping_suffix(self) -> str:
        """
        The string to add to the end of an identifier when stropping.
        """
        return self._config.get(self._section, 'stropping_suffix')

    @property
    def support_namespace(self) -> typing.List[str]:
        """
        The hierarchial namespace used by the support headers. The property
        is a dot separated string when specified in configuration. This
        property returns that value split into namespace components with the
        first identifier being the first index in the array, etc.

        .. invisible-code-block: python

            from nunavut.lang import Language
            import configparser

            config = configparser.ConfigParser()
            config.add_section('nunavut.lang.cpp')


            lang_cpp = Language('cpp', config, True)

        .. code-block:: python

            config.set('nunavut.lang.cpp', 'support_namespace', 'foo.bar')
            assert len(lang_cpp.support_namespace) == 2
            assert lang_cpp.support_namespace[0] == 'foo'
            assert lang_cpp.support_namespace[1] == 'bar'

        """
        namespace_str = self._config.get(self._section, 'support_namespace', fallback='')
        return namespace_str.split('.')

    @property
    def encoding_prefix(self) -> str:
        """
        The string to prepend to an encoding hex value.
        """
        return self._config.get(self._section, 'encoding_prefix')

    @property
    def enable_stropping(self) -> bool:
        """
        Whether or not to strop identifiers for this language.
        """
        return self._config.getboolean(self._section, 'enable_stropping')

    @property
    def has_standard_namespace_files(self) -> bool:
        """
        Whether or not the language defines special namespace files as part of
        its core standard (e.g. python's __init__).
        """
        return self._config.getboolean(self._section, 'has_standard_namespace_files')

    @property
    def omit_serialization_support(self) -> bool:
        """
        If True then generators should not include serialization routines, types,
        or support libraries for this language.
        """
        return self._omit_serialization_support

    @property
    def module(self) -> typing.Any:
        if self._module is None:
            import importlib
            self._module = importlib.import_module('nunavut.lang.{}'.format(self._language_name))
        return self._module

    def get_config_value(self,
                         key: str,
                         default_value: typing.Union[typing.Mapping[str, typing.Any], str, None] = None)\
            -> typing.Union[typing.Mapping[str, typing.Any], str, None]:
        """
        Get an optional language property from the language configuration.

        :param str key: The config value to retrieve.
        :param default_value: The value to return if the key was not in the configuration. If provided
            this method will not raise.
        :type default_value: typing.Union[typing.Mapping[str, typing.Any], str, None]
        :return: Either the value from the config or the default_value if provided.
        :rtype: typing.Union[typing.Mapping[str, typing.Any], str, None]
        :raises: KeyError if the key does not exist and a default_value was not provied.
        """
        return self._config.get(self._section, key, fallback=default_value)

    def get_config_value_as_bool(self, key: str, default_value: bool = False) -> bool:
        """
        Get an optional language property from the language configuration returning a boolean. The rules
        for boolean conversion are as follows:

        .. invisible-code-block: python

            from nunavut.lang import Language
            import configparser

            config = configparser.ConfigParser()
            config.add_section('nunavut.lang.cpp')

            lang_cpp = Language('cpp', config, True)

        .. code-block:: python

            # "Any string" = True
            config.set('nunavut.lang.cpp', 'v', 'Any string')
            assert lang_cpp.get_config_value_as_bool('v')

            # "true" = True
            config.set('nunavut.lang.cpp', 'v', 'true')
            assert lang_cpp.get_config_value_as_bool('v')

            # "TrUe" = True
            config.set('nunavut.lang.cpp', 'v', 'TrUe')
            assert lang_cpp.get_config_value_as_bool('v')

            # "1" = True
            config.set('nunavut.lang.cpp', 'v', '1')
            assert lang_cpp.get_config_value_as_bool('v')

            # "false" = False
            config.set('nunavut.lang.cpp', 'v', 'false')
            assert not lang_cpp.get_config_value_as_bool('v')

            # "FaLse" = False
            config.set('nunavut.lang.cpp', 'v', 'FaLse')
            assert not lang_cpp.get_config_value_as_bool('v')

            # "0" = False
            config.set('nunavut.lang.cpp', 'v', '0')
            assert not lang_cpp.get_config_value_as_bool('v')

            # "" = False
            config.set('nunavut.lang.cpp', 'v', '')
            assert not lang_cpp.get_config_value_as_bool('v')

            # False if not defined
            assert not lang_cpp.get_config_value_as_bool('not_a_key')

            # True if not defined but default_value is True
            assert lang_cpp.get_config_value_as_bool('not_a_key', True)

        :param str key: The config value to retrieve.
        :param bool default_value: The value to use if no value existed.
        :return: The config value as either True or False.
        :rtype: bool
        """
        result = self._config.get(self._section, key, fallback='' if not default_value else '1')
        if result.lower() == 'false' or result == '0':
            return False
        else:
            return bool(result)

    def get_templates_package_name(self) -> str:
        """
        The name of the nunavut python package containing filters, types, and configuration
        for this language.
        """
        return 'nunavut.lang.{}'.format(self.name)

    def get_named_types(self) -> typing.Mapping[str, str]:
        """
        Get a map of named types to the type name to emit for this language.
        """
        return self._config.getdict(self._section, 'named_types', fallback={})  # type: ignore

    def get_reserved_identifiers(self) -> typing.List[str]:
        """
        Get a list of identifiers that are reserved keywords for this language.
        """
        return self._config.getlist(self._section, 'reserved_identifiers', fallback=[])  # type: ignore

    def get_filters(self, make_implicit: bool = True) -> typing.Mapping[str, typing.Callable]:
        """
        Inspect the language module for functions with a name starting with "filter\\_" and return
        a map of filter names to the filter callable.

        :param bool make_implicit: If True then the function name will not contain a language prefix
            otherwise a prefix will be provided for this langauge separated by a '.'.
        :returns: A mapping of filter names to filter functions.
        """
        if make_implicit:
            return self._filters
        else:
            explicit_filters = dict()
            for key, value in self._filters.items():
                explicit_filters['{}.{}'.format(self.name, key)] = value
            return typing.cast(typing.Mapping[str, typing.Callable], explicit_filters)

    def get_globals(self, make_implicit: bool = True) -> typing.Mapping[str, typing.Any]:
        """
        Get all values for this language that should be available in a global context.

        :param bool make_implicit: If True then the global name will not contain a language prefix
            otherwise a prefix will be provided for this langauge separated by a '.'.
        :returns: A mapping of global names to global values.
        """
        globals_map = dict()
        for key, value in self.get_named_types().items():
            if make_implicit:
                globals_map['typename_{}'.format(key)] = value
            else:
                globals_map['{}.typename_{}'.format(self._language_name, key)] = value
        return typing.cast(typing.Mapping[str, typing.Callable], globals_map)


class LanguageContext:
    """
    Context object containing the current target language (if any) and used to access
    :class:`Language` objects.

    :param str target_language: If provided a :class:`Language` object will be created to hold
                                the target language set for this context. If None then there is
                                no target language.
    :param str extension: The extension to use for generated file types or None to use a default
                          based on the target_language.
    :param str namespace_output_stem: The filename stem to give to Namespace output files if
                                      emitted or None to use a default based on a target_language.
    :param additional_config_files: A list of paths to additional ini files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.ini` file and files
        appearing later in this list will override value found in earlier entries.
    :type additional_config_files: typing.List[pathlib.Path]
    :param bool omit_serialization_support_for_target: If True then generators should not include
        serialization routines, types, or support libraries for the target language.
    :raises ValueError: If extension is None and no target language was provided.
    :raises KeyError: If the target language is not known.
    """

    @classmethod
    def _as_dict(cls, value: str) -> typing.Dict[str, typing.Any]:
        value_as_dict = dict()
        value_pairs = value.strip().split('\n')
        for pair in value_pairs:
            kv = pair.strip().split('=')
            value_as_dict[kv[0].strip()] = kv[1].strip()
        return value_as_dict

    @classmethod
    def _as_list(cls, value: str) -> typing.List[str]:
        return [r.strip() for r in value.split('\n')]

    @classmethod
    def _load_config(cls, *additional_config_files: pathlib.Path) -> configparser.ConfigParser:
        import pkg_resources
        parser = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation(),
                                           converters={'dict': cls._as_dict,
                                                       'list': cls._as_list})
        resources = [r for r in pkg_resources.resource_listdir(__name__, '.') if r.endswith('.ini')]
        for resource in resources:
            with pkg_resources.resource_stream(__name__, resource) as resource_stream:
                ini_string = resource_stream.read().decode(encoding='utf-8', errors='replace')
                parser.read_string(ini_string, resource)
        for additional_path in additional_config_files:
            with open(str(additional_path), 'r') as additional_file:
                parser.read_file(additional_file)
        return parser

    def __init__(self,
                 target_language: typing.Optional[str] = None,
                 extension: typing.Optional[str] = None,
                 namespace_output_stem: typing.Optional[str] = None,
                 additional_config_files: typing.List[pathlib.Path] = [],
                 omit_serialization_support_for_target: bool = True):
        self._extension = extension
        self._namespace_output_stem = namespace_output_stem
        self._config = self._load_config(*additional_config_files)

        # create target language, if there is one.
        if target_language is None:
            self._target_language = None
        else:
            try:
                self._target_language = Language(target_language, self._config,
                                                 omit_serialization_support_for_target)
            except ImportError:
                raise KeyError('{} is not a supported language'.format(target_language))
            if namespace_output_stem is not None:
                self._config.set('nunavut.lang.{}'.format(target_language),
                                 'namespace_file_stem',
                                 namespace_output_stem)
            if extension is not None:
                self._config.set('nunavut.lang.{}'.format(target_language),
                                 'extension',
                                 extension)

        # create remaining languages
        self._languages = dict()  # type: typing.Dict[str, Language]
        for language_name in self.get_supported_language_names():
            if self._target_language is not None and self._target_language.name == language_name:
                self._languages[language_name] = self._target_language
            else:
                try:
                    self._languages[language_name] = Language(language_name, self._config, False)
                except ImportError:
                    raise KeyError('{} is not a supported language'.format(language_name))

    def get_language(self, key_or_modulename: str) -> Language:
        """
        Get a :class:`Language` object for a given language identifier.

        :param str key_or_modulename: Either one of the Nunavut mnemonics for a supported language or
            the ``__name__`` of one of the ``nunavut.lang.[language]`` python modules.
        :returns: A :class:`Language` object cached by this context.
        :rtype: Language
        """
        if key_or_modulename is None or len(key_or_modulename) == 0:
            raise ValueError('key argument is required.')
        key = (key_or_modulename[13:] if key_or_modulename.startswith('nunavut.lang.') else key_or_modulename)
        return self.get_supported_languages()[key]

    def get_supported_language_names(self) -> typing.Iterable[str]:
        """Get a list of target languages supported by Nunavut.

        :returns: An iterable of strings which are languages with special
            support within Nunavut templates.
        """
        return [s[13:] for s in self._config.sections() if s.startswith('nunavut.lang.')]

    def get_output_extension(self) -> str:
        """
        Gets the output extension to use regardless of a target language being available or not.

        :returns: A file extension name with a leading dot.
        """
        if self._extension is not None:
            return self._extension
        elif self._target_language is not None:
            return self._target_language.extension
        else:
            raise RuntimeError('No extension was provided and no target language was set.'
                               'Cannot determine the extension to use.')

    def get_default_namespace_output_stem(self) -> typing.Optional[str]:
        """
        The filename stem to give to Namespace output files if emitted or None if there was none
        specified and there is no target language.

        :returns: A file name stem or None
        """
        return self._namespace_output_stem

    def get_target_language(self) -> typing.Optional[Language]:
        """
        Returns the target language configured on this object or None
        if no target language was specified.
        """
        return self._target_language

    def get_target_id_filter(self) -> typing.Callable[[str], str]:
        """
        A filter that will transform a given string into a valid identifier
        in the target language. The string is pass through unmodified if no
        target language was set.
        """
        if self._target_language is not None:
            filters = self._target_language.get_filters()
            for name, filter in filters.items():
                if name == 'id':
                    id_filter = typing.cast(typing.Callable[[Language, str], str], filter)
                    return functools.partial(id_filter, self._target_language)

        return lambda unfiltered: unfiltered

    def get_supported_languages(self) -> typing.Dict[str, Language]:
        """
        Returns a collection of available language support objects.
        """
        return self._languages

    @property
    def config(self) -> configparser.ConfigParser:
        return self._config


class _UniqueNameGenerator:
    """
    Functor used by template filters to obtain a unique name within a given template.
    This should be made available as a private global within each template.
    """
    _singleton = None  # type: typing.Optional['_UniqueNameGenerator']

    def __init__(self) -> None:
        self._index_map = {}  # type: typing.Dict[str, typing.Dict[str, int]]

    @classmethod
    def reset(cls) -> None:
        cls._singleton = cls()

    @classmethod
    def get_instance(cls) -> '_UniqueNameGenerator':
        if cls._singleton is None:
            raise RuntimeError('No _UniqueNameGenerator has been created. Please use reset to create.')
        return cls._singleton

    def __call__(self, key: str, base_token: str, prefix: str, suffix: str) -> str:
        """
        Uses a global index to generate a number unique to a given base_token within a template
        for a given domain (key).
        """
        try:
            keymap = self._index_map[key]
        except KeyError:
            keymap = {}
            self._index_map[key] = keymap

        try:
            next_index = keymap[base_token]
            keymap[base_token] = next_index + 1
        except KeyError:
            next_index = 0
            keymap[base_token] = 1

        return "{prefix}{base_token}{index}{suffix}".format(
            prefix=prefix,
            base_token=base_token,
            index=next_index,
            suffix=suffix)
